// Copyright 2022-2025 David Robillard <d@drobilla.net>
// SPDX-License-Identifier: ISC

#include <lilv/lilv.h> // IWYU pragma: keep

#ifdef __GNUC__
__attribute__((const))
#endif
int
main(void)
{
  return 0;
}
